/* $Id: s270ctrl.c 95 2006-02-13 21:04:19Z lennart $ */

/***
  This file is part of s270ctrl.
  
  s270ctrl is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  s270ctrl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with s270ctrl; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
  02110-1301, USA.
***/

#define _GNU_SOURCE

#include <time.h>
#include <unistd.h>
#include <sys/io.h> 
#include <inttypes.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#define MSI_PORT_CONTROL 0x66
#define MSI_PORT_DATA 0x62
#define MSI_PORT_AUX 0x90

#define MSI_COMMAND_LCD_LEVEL 0x11

#define MSI_LCD_LEVEL_MAX 9
#define MSI_LCD_LEVEL_INVALID 0x90

static void msleep(unsigned msec) {
    struct timespec ts;

    ts.tv_sec = msec / 1000;
    ts.tv_nsec = 1000000 * (msec % 1000);
    
    nanosleep(&ts, NULL);
}

int kbc_wait_for_0x10(int msec) {

    while (msec-- > 0) {

        if (!(inb(MSI_PORT_CONTROL) & 0x10))
            return 0;
        
        msleep(1);
    }

    return -1;
}

int kbc_wait_for_0x02(int msec) {

    while (msec-- > 0) {
        
        if (!(inb(MSI_PORT_CONTROL) & 2))
            return 0;
        
        msleep(1);
    }

    return -1;
}

int kbc_wait_for_0x01(int msec) {

    while (msec-- > 0) {
        inb(MSI_PORT_DATA);

        if (!(inb(MSI_PORT_CONTROL) & 1))
            return 0;
        
        msleep(1); 
    }

    return -1;
}

int kbc_write_0x84_wait_for_0x10(int msec) {

    while (msec-- > 0) {
        outb(0x84, MSI_PORT_CONTROL);

        if (!(inb(MSI_PORT_DATA) & 10))
            return 0;
        
        msleep(1);
    }

    return -1;
}

void kbc_reset(int full) {
    if (full) {
        outb(0x82, MSI_PORT_CONTROL);
        inb(MSI_PORT_AUX);
    } else
        outb(0x83, MSI_PORT_CONTROL);
}

int kbc_access(char cmd, const uint8_t*wdata, uint8_t*rdata, int wlen, int rlen, int msec) {
    if (kbc_write_0x84_wait_for_0x10(msec) < 0)
        return -1;

    if (kbc_wait_for_0x01(msec) < 0) 
        goto fail;

    if (kbc_wait_for_0x10(msec) < 0)
        goto fail;

    kbc_reset(1);

    if (kbc_wait_for_0x02(msec) < 0)
        goto fail;

    outb(cmd, MSI_PORT_CONTROL);

    for (; wlen > 0; wlen--) {
        if (kbc_wait_for_0x02(msec) < 0)
            goto fail;

        outb(*(wdata++), MSI_PORT_DATA);
    }

     msleep(100); 
    
    if (rlen) {
        int i;

        for (i = 0; i < msec; i++) { 
            *rdata = inb(MSI_PORT_DATA);
            if (*rdata != MSI_LCD_LEVEL_INVALID)  
                break; 

            msleep(1);
        }

        if (i >= msec)
            goto fail;
    }

    if (kbc_wait_for_0x02(msec) < 0)
        goto fail;
    
    kbc_reset(0);

    msleep(100);

    return 0;
    
fail:
    kbc_reset(0);
    
    return -1;
}

static const uint8_t lcd_table[MSI_LCD_LEVEL_MAX] = {
    0x00, 0x1f, 0x3e, 0x5d, 0x7c, 0x9b, 0xba, 0xd9, 0xf8
};

static int set_lcd_level(int level) {
    uint8_t buf[2] = { 0x80,  0x00 };

    if (level < 0 || level >= MSI_LCD_LEVEL_MAX)
        return -1;

    buf[1] = lcd_table[level];
/*     printf("setting to: 0x%02x\n", buf[1]);  */
    
    return kbc_access(MSI_COMMAND_LCD_LEVEL, buf, NULL, 2, 0, 300);
}

static int get_lcd_level(void) { 
    const uint8_t wdata[1] = { 0 };
    uint8_t rdata[1] = { 0 };
    int i;

    if (kbc_access(MSI_COMMAND_LCD_LEVEL, wdata, rdata, 1, 1, 300) < 0) 
        return -1;

/*     printf("read: 0x%02x\n", rdata[0]);  */

    for (i = 0; i < MSI_LCD_LEVEL_MAX; i++)
        if (lcd_table[i] == rdata[0])
            return i;
    
    return 0;
}
    
int main(int argc, char *argv[]) {
    int ret = 1;
    char *e;

    if ((e = strrchr(argv[0], '/')))
        argv[0] = e+1;
    
    if (argc > 2) {
        fprintf(stderr, "Too many arguments.\n");
        goto finish;
    }

    if (argc > 1 && (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0)) {
        printf("%s - Set LCD brightness of MSI S270 laptops (aka Cytron/TCM/Medion/Tchibo MD96100, SAM2000)\n\n"
               "Set LCD brightness:\n"
               "\t%s [0-8]\n\n"
               "Query LCD brightness:\n"
               "\t%s\n\n", argv[0], argv[0], argv[0]);
        ret = 0;
        goto finish;
    }
    
    if (ioperm(MSI_PORT_DATA, 1, 1) < 0 ||
        ioperm(MSI_PORT_CONTROL, 1, 1) < 0 ||
        ioperm(MSI_PORT_AUX, 1, 1) < 0) {
        fprintf(stderr, "ioperm(): %s\n", strerror(errno));
        goto finish;
    }

    if (argc > 1) {
        int level;
        
        errno = 0;
        level = strtol(argv[1], NULL, 0);
        
        if (errno != 0) {
            fprintf(stderr, "Failed to parse <%i>.\n", level);
            goto finish;
        }

        if (set_lcd_level(level) < 0) {
            fprintf(stderr, "set_lcd_level() failed\n");
            goto finish;
        }
        
    } else {
        int level;
        
        if ((level = get_lcd_level()) < 0) {
            fprintf(stderr, "get_lcd_level() failed.\n");
            goto finish;
        }

        printf("%i\n", level);
    }

    ret = 0;
    
finish:
    
    return ret;
}
